import unittest

import numpy as np

import quantum_inferno.synth.benchmark_signals as bs


class TestBenchmarkSignals(unittest.TestCase):
    def test_signal_gate(self):
        self.assertEqual(1, 1)
